/* Copyright (c) 2007 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/** 
 * CV.cpp 
 * @author James Geraci
 * Implements the equations COP, MB, EK which control
 * the change of the concentration, phil, and phis with time
 * also sets up the Jacobian
 *
 */
#include "StdAfx.h"
#include "CV.h"

CV::CV(void)
{
	myFV =(FV*) NULL;
}

CV::~CV(void)
{
}

double CV::returnmbScale(void){
	return myFV->mbScale;
};

void CV::setPosition(int x){
	myPosition = x;
}

void CV::computedxSize(void){
	mydx = (myFV->returndx())/2;
	mydx2 = 2*mydx;
	invmydx = 1/mydx;
	invmydx2 = 1/mydx2;
	mydxsqrd = mydx*mydx;
	return;
}

void CV::computedySize(void){
	mydy = (myFV->returndy())/2;
	mydy2 = 2*mydy;
	invmydy = 1/mydy;
	invmydy2 = 1/mydy2;
	mydysqrd = mydy*mydy;
	return;
}

void CV::computedzSize(void){
	mydz = (myFV->returndz());
	mydz2 = 2*mydz;
	invmydz = 1/mydz;
	invmydz2 = 1/mydz2;
	mydzsqrd = mydz*mydz;
	return;
}

void CV::computemyVolume(void){
	myVolume = mydx*mydy*mydz;
	myFV->setCVvolume(myPosition,myVolume);
	return;
}

double CV::returnmyVolume(void){
	return myVolume;
}

void CV::setpvVolume(double vol){
	pvVolume = vol;
	myPVvolumeFraction = myVolume/pvVolume;
	myFV->setCVvolumeRatio(myPosition,myPVvolumeFraction);
	return;
}
/*************************************************************/
/*******************  Work for EK ****************************/
/*************************************************************/
double CV::setekCV(void){
	overp = (Vptr[phisc] - Vptr[philc] - myFV->phi_eq);
	gls = ((myFV->als)*N*F)/(R*myFV->T);
	gsl = ((myFV->asl)*N*F)/(R*myFV->T);
	ekCV = -myArea()*(myFV->io)*(Vptr[hpc]/myFV->hnot)*expTerm()*mydx*mydy;
	return ekCV;
};

double CV::myArea(void){
	if(I >= 0)  // discharge
	{
		Area = (myFV->Amax)*(myFV->soceta());
	}
	else{
		Area = (myFV->Amax)*(1-myFV->soceta());
	}
	
	return Area;
};

double CV::expTerm(void){
	myexpTerm = exp(gsl*overp) - exp(-gls*overp);
	return myexpTerm;
};

double CV::ekJeps(void){
	return 0;
};

double CV::ekJhp(void){
	return -myArea()*((myFV->io)/(myFV->hnot))*myexpTerm*mydx*mydy;
};

double CV::ekJphis(void){
	double phisderiv = -myArea()*myFV->io*(Vptr[hpc]/myFV->hnot)*(gsl*exp(gsl*overp)+gls*exp(-gls*overp))*mydx*mydy;
	return phisderiv;
};

double CV::ekJphil(void){
	double philderiv = myArea()*myFV->io*(Vptr[hpc]/myFV->hnot)*(gsl*exp(gsl*overp)+gls*exp(-gls*overp))*mydx*mydy;
	return philderiv;
};

/*************************************************************/
/*******************  Work for DOC ***************************/
/*************************************************************/

double CV::setdocCV(void){
	docCV = setdocCVjl() + setdocCVse();
	return docCV;
};

double CV::inCurrentCV(void){
	return mydx*I;
};

double CV::setdocCVjl(void){
	return myFjlx()*mydy + myFjly()*mydx;
};

double CV::setdocCVse(void){
	return myFsex()*mydy + myFsey()*mydx;
};

double CV::myFjly(void){
	return myYnorm*myFV->returnFjly(myPosition);
};

double CV::myFjlx(void){
	return myXnorm*myFV->returnFjlx(myPosition);
};

double CV::myFsey(void){
	return myYnorm*myFV->returnFsey(myPosition);
};

double CV::myFsex(void){
	return myXnorm*myFV->returnFsex(myPosition);
};

double CV::myJphil(void){
	return mydy*(myFV->returnKeffx()/(myFV->returndx()))
		+ mydx*myFV->returnKeffy()/(myFV->returndy());
};

// Phil Jacobian Contributions
double CV::myJxphil(void){
	return -mydy*myFV->returnKeffx()/(myFV->returndx());
};

double CV::myJyphil(void){
	return -mydx*myFV->returnKeffy()/(myFV->returndy());
};
// HP Jacobian Contributions
double CV::myDOCJhp(void){
	return mydy*(myFV->returnolsKscalex()/(myFV->returndx()))
		+ mydx*myFV->returnolsKscaley()/(myFV->returndy());
//	return 0;
};

double CV::myDOCJxhp(void){
	return -mydy*myFV->returnolsKscalex()/(myFV->returndx());
//	return 0;
};

double CV::myDOCJyhp(void){
	return -mydx*myFV->returnolsKscaley()/(myFV->returndy());
};
// Phis Jacobian Contributions
double CV::myJphis(void){
	return mydy*(myFV->returnsigeffx()/(myFV->returndx()))
		+ mydx*myFV->returnsigeffy()/(myFV->returndy());
};

double CV::myJxphis(void){
	return -mydy*myFV->returnsigeffx()/(myFV->returndx());
};

double CV::myJyphis(void){
	return -mydx*myFV->returnsigeffy()/(myFV->returndy());
};

/*************************************************************/
/*******************  Time Derivatives for MB ****************/
/*************************************************************/
double CV::setdhdt(void)
{
	dhdt = (Vptr[hpc] - pVptr[hpc])/dt;
	return dhdt;
};

double CV::setdhdtCV(void){
	dhdtCV = myFV->returneps()*dhdt*mydx*mydy + myFdx()*mydy + myFdy()*mydx;
	return dhdtCV;
};

double CV::myJhp(void){
	return mydx*mydy*myFV->returneps()/dt + mydy*(myFV->returnDeffx()/(myFV->returndx()))
		 + mydx*myFV->returnDeffy()/(myFV->returndy());
};

double CV::myJxhp(void){
	return -mydy*myFV->returnDeffx()/(myFV->returndx());
};

double CV::myJyhp(void){
	return -mydx*myFV->returnDeffy()/(myFV->returndy());
};

double CV::myFdx(void){
	return myXnorm*myFV->returnFdx(myPosition);
};

double CV::myFdy(void){
	return myYnorm*myFV->returnFdy(myPosition);
};
