/* Copyright (c) 2007 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/* 
 * @author James Geraci
 * object that contains the description of the FV volumes
 * these objects control all the fluxes within the model
 * and the evolution of the porosity
 */
#include "FV.h"
#include "StdAfx.h"

FV::FV(void)
{
}

FV::FV(volumeSpatialData* sData, matrixData *mD,electrodeSizeData* esData, volumeChemData* vcData, electrodeChemData* ecData, volumeIConditions* iData)
{
	initializeMatrix(mD);
	initializePosition(sData,mD,esData);
	initializeOtherPositions();
	initializeSize(sData);
	initializeElectrodeChem(ecData);
	initializeVolumeChem(vcData);
	setInitialConditions(iData);
	return;
}

FV::~FV(void)
{
}

void FV::initializeMatrix(matrixData *mData)
{
	fluxVptr = mData->myfluxV();
	epsVptr = mData->myepsV();
	socVptr = mData->mysocV();
	pVptr = mData->mypsV();
	Vptr = mData->mysV();
	iVptr = mData->myiV();
	divVptr = mData->mydivV();
	Jacobptr = mData->myJacobian(); 
	numPVColumns = mData->myPVColumns();
	numPVRows = mData->myPVRows();
	numPVZrows = mData->myPVZrows();
	totalPVvolumes = numPVColumns*numPVRows*numPVZrows;
	
	numFVColumns = mData->myFVColumns();
	numFVRows = mData->myFVRows();
	numFVZrows = mData->myFVZrows();
	totalFVvolumes = numFVColumns*numFVRows*numFVZrows;
	
	return;
}

void FV::initializePosition(volumeSpatialData *sData, matrixData* mData, electrodeSizeData* eData)
{
		// Cell Position Data Initialization
	myColumn = sData->getColumn()+eData->getFVstartColumn();
	myRow = sData->getRow();
	myZrow = sData->getZrow();
//	rowLength = mData->mysvLength();

 	presV = myColumn + myRow*mData->myFVColumns();
//	pprevV = presV-2;
	prevV = presV-1;
	nextV = presV+1;
	//nnextV = presV+2;
	
	topV = presV - mData->myFVColumns();
	lowerV = presV + mData->myFVColumns();
	return;
}

void FV::initializeOtherPositions(void)
{
	return;
};

void FV::initializeSize(volumeSpatialData *sData)
{
	// Cell Size Data Initialization
	dx = sData->getxWidth();
	dy = sData->getyHeight();
	dz = sData->getzDepth();

	myVolume = dx*dy*dz;

	dx2 = 2*dx;
	invdx = 1/dx;
	invdx2 = 1/dx2;
	dxsqrd = dx*dx;

	dy2 = 2*dy;
	invdy = 1/dy;
	invdy2 = 1/dy2;
	dysqrd = dy*dy;
	return;	
}

void FV::initializeElectrodeChem(electrodeChemData *cData)
{
	tnp = cData->gettnp();
	mbScale = cData->getmbScale();
	olsScale = cData->getolsScale();
	myVf = cData->getmyVf();
	myVfd2 = myVf/(2*F);
	io = cData->getio();
	return;
};

void FV::initializeVolumeChem(volumeChemData *vData)
{
	T = vData->getT();
	Amax = vData->getAmax();
	Qmax = vData->getQmax();
	soc = vData->getSoc();
	eta = vData->getEta();
	als = vData->getals();
	asl = vData->getasl();
	gc = (als*R*T*N)/(F);
	ga = (asl*R*T*N)/(F);
	Dx = vData->getDx();
	Dy = vData->getDy();
	sigx = vData->getsigx();
	sigy = vData->getsigy();
	Kx = vData->getKx();
	Ky = vData->getKy();
	k = vData->getk();
	m = vData->getm();
	exm = vData->getexm();
	ex = vData->getex();
	hnot = vData->gethnot();
	epsnot = vData->getepsnot();
	mySign = vData->getmySign();
	phi_eq = vData->getphi_eq();
	return;
}


void FV::setupPvPointers(void){
  if(mode == 0){
    offSet = totalPVvolumes;
    numVars = 1;
  }else{
    offSet = 1;
    numVars = myVars;
  }
  
  setULPV();
  setURPV();
  setLLPV();
  setLRPV();
  
  uFjlx = (presV)*fluxesPerVar;
  bFjlx = uFjlx+1;
  lFjly = bFjlx+1;
  rFjly = lFjly+1;
  
  uFdx = (presV+totalFVvolumes)*fluxesPerVar;
  bFdx = uFdx+1;
  lFdy = bFdx+1;
  rFdy = lFdy+1;
  
  uFsex = (presV+2*totalFVvolumes)*fluxesPerVar;
  bFsex = uFsex+1;
  lFsey = bFsex+1;
  rFsey = lFsey+1;
  
  uFlncx = (presV+3*totalFVvolumes)*fluxesPerVar;
  bFlncx = uFlncx+1;
  lFlncy = bFlncx+1;
  rFlncy = lFlncy+1;
  
  uFlex = (presV+4*totalFVvolumes)*fluxesPerVar;
  bFlex = uFlex+1;
  lFley = bFlex+1;
  rFley = lFley+1;
  return;
};

void FV::setULPV(void){
	ulPVdiv = myRow*(numPVColumns) + myColumn;
	ulPVhpc = numVars*(myRow*(numPVColumns) + myColumn);
	ulPVphilc = ulPVhpc + offSet;
	ulPVphisc = ulPVhpc + 2*offSet;
	return;
};

void FV::setURPV(void){
	urPVdiv = myRow*(numPVColumns) + myColumn + 1;
	urPVhpc = numVars*(myRow*(numPVColumns) + myColumn + 1);
	urPVphilc = urPVhpc + offSet;
	urPVphisc = urPVhpc + 2*offSet;
	return;
};

void FV::setLLPV(void){
	llPVdiv = (myRow+1)*(numPVColumns) + myColumn;
	llPVhpc = numVars*((myRow+1)*(numPVColumns) + myColumn);
	llPVphilc = llPVhpc + offSet;
	llPVphisc = llPVhpc + 2*offSet;
	return;
};

void FV::setLRPV(void){
	lrPVdiv = (myRow+1)*(numPVColumns) + myColumn + 1;
	lrPVhpc = numVars*((myRow+1)*(numPVColumns) + myColumn + 1);
	lrPVphilc = lrPVhpc + offSet;
	lrPVphisc = lrPVhpc + 2*offSet;
	return;
};
/*
void FV::setURPV(void){
  urPVdiv = myRow*(numPVColumns) + myColumn + 1;
  urPVhpc = myRow*(numPVColumns) + myColumn + 1;
  urPVphilc = urPVhpc + totalPVvolumes;
  urPVphisc = urPVhpc + 2*totalPVvolumes;
  return;
};

void FV::setULPV(void){
  ulPVdiv = myRow*(numPVColumns) + myColumn;
  ulPVhpc = myRow*(numPVColumns) + myColumn;
  ulPVphilc = ulPVhpc + totalPVvolumes;
  ulPVphisc = ulPVhpc + 2*totalPVvolumes;
  return;
};

void FV::setLLPV(void){
  llPVdiv = (myRow+1)*(numPVColumns) + myColumn;
  llPVhpc = (myRow+1)*(numPVColumns) + myColumn;
  llPVphilc = llPVhpc + totalPVvolumes;
  llPVphisc = llPVhpc + 2*totalPVvolumes;
	return;
};

void FV::setLRPV(void){
  lrPVdiv = (myRow+1)*(numPVColumns) + myColumn + 1;
  lrPVhpc = (myRow+1)*(numPVColumns) + myColumn + 1;
  lrPVphilc = lrPVhpc + totalPVvolumes;
  lrPVphisc = lrPVhpc + 2*totalPVvolumes;
  return;
};
*/

void FV::setInitialConditions(volumeIConditions* iData)
{
	eps = iData->geteps();
	epsVptr[presV] = eps;
	socVptr[presV] = soc;
//	soc = socVptr[presV];
//	Vptr[hpc] = iData->gethp();
//	Vptr[phi_lc] = iData->getphi_l();
//	Vptr[phi_sc] = iData->getphi_s();
//	Vptr[jlxc] = iData->getjlx();
	return;
}


void FV::setMaterialPropertyEffectiveValue(double *EffectivePropertyValue, double PropertyValue, double tortuosityFactor){
	*EffectivePropertyValue = PropertyValue*pow(eps,tortuosityFactor);
return;
};

void FV::setMaterialPropertyPermeability(){
	invKp = (180*pow((1-eps),2))/(pow(eps,3)*pow(d,2));
	return;
};

void FV::computeFlux(double *Flux, double EMPV, double V2, double V1, double spacing){
	*Flux = -EMPV*(V2 - V1)/spacing;
	return;
};

void FV::computeCurrent(double *Fjl, double Fle, double Flc){
	*Fjl = Fle + Flc;
	return;
};

void FV::computeFluxes(void){
	// Compute Diffusion Fluxes
	computeFlux(&fluxVptr[uFdx],Deffx,Vptr[urPVhpc],Vptr[ulPVhpc],dx);
	computeFlux(&fluxVptr[bFdx],Deffx,Vptr[lrPVhpc],Vptr[llPVhpc],dx);
	computeFlux(&fluxVptr[lFdy],Deffy,Vptr[ulPVhpc],Vptr[llPVhpc],dy);
	computeFlux(&fluxVptr[rFdy],Deffy,Vptr[urPVhpc],Vptr[lrPVhpc],dy);

	// Compute Chemical Potential Fluxes
	computeFlux(&fluxVptr[uFlncx],olsKscalex,log(Vptr[urPVhpc]),log(Vptr[ulPVhpc]),dx);
	computeFlux(&fluxVptr[bFlncx],olsKscalex,log(Vptr[lrPVhpc]),log(Vptr[llPVhpc]),dx);
	computeFlux(&fluxVptr[lFlncy],olsKscaley,log(Vptr[ulPVhpc]),log(Vptr[llPVhpc]),dy);
	computeFlux(&fluxVptr[rFlncy],olsKscaley,log(Vptr[urPVhpc]),log(Vptr[lrPVhpc]),dy);

	// Compute Liquid Current Fluxes
	computeFlux(&fluxVptr[uFlex],Keffx,Vptr[urPVphilc],Vptr[ulPVphilc],dx);
	computeFlux(&fluxVptr[bFlex],Keffx,Vptr[lrPVphilc],Vptr[llPVphilc],dx);
	computeFlux(&fluxVptr[lFley],Keffy,Vptr[ulPVphilc],Vptr[llPVphilc],dy);
	computeFlux(&fluxVptr[rFley],Keffy,Vptr[urPVphilc],Vptr[lrPVphilc],dy);
 /*
	computeFlux(&fluxVptr[uFjlx],Keffx,Vptr[urPVphilc],Vptr[ulPVphilc],dx);
	computeFlux(&fluxVptr[bFjlx],Keffx,Vptr[lrPVphilc],Vptr[llPVphilc],dx);
	computeFlux(&fluxVptr[lFjly],Keffy,Vptr[ulPVphilc],Vptr[llPVphilc],dy);
	computeFlux(&fluxVptr[rFjly],Keffy,Vptr[urPVphilc],Vptr[lrPVphilc],dy);
*/
	// Compute Solid Current Fluxes
	computeFlux(&fluxVptr[uFsex],sigeffx,Vptr[urPVphisc],Vptr[ulPVphisc],dx);
	computeFlux(&fluxVptr[bFsex],sigeffx,Vptr[lrPVphisc],Vptr[llPVphisc],dx);
	computeFlux(&fluxVptr[lFsey],sigeffy,Vptr[ulPVphisc],Vptr[llPVphisc],dy);
	computeFlux(&fluxVptr[rFsey],sigeffy,Vptr[urPVphisc],Vptr[lrPVphisc],dy);

	// Compute the Liquid Phase ElectroChemical Fluxes
	computeCurrent(&fluxVptr[uFjlx],fluxVptr[uFlex],fluxVptr[uFlncx]);
	computeCurrent(&fluxVptr[bFjlx],fluxVptr[bFlex],fluxVptr[bFlncx]);
	computeCurrent(&fluxVptr[lFjly],fluxVptr[lFley],fluxVptr[lFlncy]);
	computeCurrent(&fluxVptr[rFjly],fluxVptr[rFley],fluxVptr[rFlncy]);
	return;
};

void FV::computeMaterialProperties(void){
	setMaterialPropertyEffectiveValue(&Deffx,Dx,ex);
	setMaterialPropertyEffectiveValue(&Deffy,Dy,ex);
	setMaterialPropertyEffectiveValue(&Keffx,Kx,ex);
	setMaterialPropertyEffectiveValue(&Keffy,Ky,ex);
	setMaterialPropertyEffectiveValue(&sigeffx,sigx,exm);
	setMaterialPropertyEffectiveValue(&sigeffy,sigy,exm);
	setMaterialPropertyPermeability();
	olsKscalex = Keffx*olsScale;
	olsKscaley = Keffy*olsScale;
	return;
};

void FV::computeDiv(void){
	myDiv = (divVptr[ulPVdiv])*myCVvolumeRatio[myUL]
		  + divVptr[urPVdiv]*myCVvolumeRatio[myUR]
		  + divVptr[llPVdiv]*myCVvolumeRatio[myLL]
		  + divVptr[lrPVdiv]*myCVvolumeRatio[myLR];
	return;
};


/***********************************/
/********** EPS ********************/
/***********************************/
void FV::computeNextEPS(void)
{
	eps = eps + dt*myVfd2*myDiv;
	epsVptr[presV] = eps;
	return;
}

/***********************************/
/********** SOC ********************/
/***********************************/
void FV::computeNextSOC(void)
{
	soc = soc + mySign*(dt*myDiv)/Qmax;
	socVptr[presV] = soc;
	return;
}


/***********************************/
/******* Diffusion Fluxes **********/
/***********************************/
double FV::returnFdx(int p){
	if(p == UL || p == UR)
		return fluxVptr[bFdx];
	else if( p == LL || p == LR)
		return fluxVptr[uFdx];
	else
	return 0;
};

double FV::returnFdy(int p){
	if(p == UL || p == LL)
		return fluxVptr[lFdy];
	else if (p == UR || p == LR)
		return fluxVptr[rFdy];
	else
	return 0;
};


/***********************************/
/* Current in Solution Fluxes ******/
/***********************************/
double FV::returnFjlx(int p){
	if(p == UL || p == UR)
		return fluxVptr[bFjlx];
	else if( p == LL || p == LR)
		return fluxVptr[uFjlx];
	else
	return 0;
};

double FV::returnFjly(int p){
	if(p == UL || p == LL)
		return fluxVptr[lFjly];
	else if (p == UR || p == LR)
		return fluxVptr[rFjly];
	else
	return 0;
};

/***********************************/
/* Current in Solid Fluxes *********/
/***********************************/
double FV::returnFsex(int p){
	if(p == UL || p == UR)
		return fluxVptr[bFsex];
	else if (p == LL || p == LR)
		return fluxVptr[uFsex];
	else
	return 0;
};

double FV::returnFsey(int p){
	if(p == UL || p == LL)
		return fluxVptr[lFsey];
	else if (p == UR || p == LR)
		return fluxVptr[rFsey];
	else
	return 0;
};

/*************************************/
/* x direction Material Coefficients */
/*************************************/
double FV::returnDeffx(void){
	return Deffx;
};

double FV::returnKeffx(void){
	return Keffx;
};

double FV::returnolsKscalex(void){
	return olsKscalex;
};

double FV::returnsigeffx(void){
	return sigeffx;
};

double FV::returndx(void){
	return dx;
};


/*************************************/
/* y direction Material Coefficients */
/*************************************/
double FV::returnDeffy(void){
	return Deffy;
};

double FV::returnKeffy(void){
	return Keffy;
};

double FV::returnolsKscaley(void){
	return olsKscaley;
};

double FV::returnsigeffy(void){
	return sigeffy;
};

double FV::returndy(void){
	return dy;
};

double FV::returndz(void){
	return dz;
};

void FV::setdx(double d){
	dx = d;
	return;
};

void FV::setCVvolume(int position, double vol){
	myCVvolume[position] = vol;
	return;
};

void FV::setCVvolumeRatio(int position, double volRatio){
	myCVvolumeRatio[position] = volRatio;
	return;
};

double FV::returneps(void){
	return eps;
};

double FV::soceta(void){
	return pow(soc,eta);
};

