/* Copyright (c) 2007 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/* @author James Geraci
 */
#ifndef _FV_H
#define _FV_H
#include "volumeSpatialData.h"
#include "StdAfx.h"

class FV
{
public:
	FV(void);
	FV(volumeSpatialData*,matrixData*,electrodeSizeData*,volumeChemData*,electrodeChemData*, volumeIConditions*);
public:
	~FV(void);

	void computeFluxes(void);
	void computeMaterialProperties(void);
	void computeDiv(void);
	void setupPvPointers(void);

	double returnDeffx(void);
	double returnDeffy(void);
	
	double returnKeffx(void);
	double returnKeffy(void);
	
	double returnolsKscalex(void);
	double returnolsKscaley(void);

	double returnsigeffx(void);
	double returnsigeffy(void);
	
	double returndx(void);
	double returndy(void);
	double returndz(void);

	double returnFdx(int);	
	double returnFdy(int);

	double returnFjlx(int);	
	double returnFsex(int);	

	double returnFjly(int);
	double returnFsey(int);

	void computeNextEPS(void);
	void computeNextSOC(void);
	double returneps(void);
	void setdx(double);
	void setCVvolume(int,double);
	void setCVvolumeRatio(int,double);

	double soceta(void);

	// Diffusion Flux in x and y directions
	int uFdx;
	int bFdx;
	int lFdy;
	int rFdy;

	int uFlex;
	int bFlex;
	int lFley;
	int rFley;

	int uFsex;
	int bFsex;
	int lFsey;
	int rFsey;

	int uFlncx;
	int bFlncx;
	int lFlncy;
	int rFlncy;

	int uFjlx;
	int bFjlx;
	int lFjly;
	int rFjly;

	double T;
	double Amax;  // Interfacial Factor
	double als; // liquid to Solid Factor
	double asl; // Solid
	double gc;
	double ga;	
	double io; 

	double Dx;
	double Dy;
	double eps;
	double soc;
	double eta;
	double sigx;
	double Kx;
	double sigy;
	double Ky;
	double invKp;	// permeability
	double d;       // mean diameter of particle
	double k;
	double m;
	double exm;
	double ex;
	
	double hnot;
	double epsnot;
	double phi_eq;
	double epsspan;
	bool charging;
	double mySign;
	double mbScale;

	double Qmax;
protected:
	// Data Matrix Pointers
	double* fluxVptr;
	double* divVptr;
	double* epsVptr;
	double* socVptr;
	double* pVptr;
	double* Vptr;
	double* iVptr;	
	double** Jacobptr;
	
	int numPVColumns;
	int numPVRows;
	int numPVZrows;
	int totalPVvolumes;

	int numFVColumns;
	int numFVRows;
	int numFVZrows;
	int totalFVvolumes;

	void setULPV(void);
	void setURPV(void);
	void setLLPV(void);
	void setLRPV(void);

	// Spatial Attributes
	double dx;
	double dy;
	double dz;
	
	double dx2;
	double invdx;
	double invdx2;
	double dxsqrd;

	double dy2;
	double invdy;
	double invdy2;
	double dysqrd;

	double myVolume;
	double myCVvolume[4];
	double myCVvolumeRatio[4];
	double myDiv;

	// physical attributes	
	double tnp;
//	double mbScale;
	double olsScale;
	double myVf;
	double myVfd2;
	

	double Deffx;
	double Deffy;
	double Keffx;
	double Keffy;
	double sigeffx;
	double sigeffy;

	double olsKscalex;
	double olsKscaley;
	
	// functions
	void initializeMatrix(matrixData*);
	void initializePosition(volumeSpatialData*, matrixData*, electrodeSizeData*);
	void initializeOtherPositions(void);
	void initializeSize(volumeSpatialData*);
	void initializeElectrodeChem(electrodeChemData*);
	void initializeVolumeChem(volumeChemData*);
	void setInitialConditions(volumeIConditions*);

	void setMaterialPropertyEffectiveValue(double*,double,double);
	void setMaterialPropertyPermeability(void);
	void computeFlux(double*,double,double,double,double);
	void computeCurrent(double*,double,double);



// Position information for Cell
	int myColumn;
	int myRow;
	int myZrow;
	int JRow;
	int JRowBuffer;

// Positioning Data
	int rowLength;

//  Used in Calculating Addresses
	int pprevV;
	int prevV;
	int presV;
	int nextV;
	int nnextV;
	
	int topV;
	int lowerV;

	int offSet;
	int numVars;

// present volume's properties
	int epsc;
	
// Upper Left PV volume's properties
	int ulPVdiv;
	int ulPVhpc;
	int ulPVphilc;
	int ulPVphisc;
// Upper Right PV volume's properties
	int urPVdiv;
	int urPVhpc;
	int urPVphilc;
	int urPVphisc;
// Lower Left PV volume's properties
	int llPVdiv;
	int llPVhpc;
	int llPVphilc;
	int llPVphisc;
// Lower Rigth PV volume's properties
	int lrPVdiv;
	int lrPVhpc;
	int lrPVphilc;
	int lrPVphisc;
// Defines
	static const int myUL = 3;
	static const int myUR = 2;
	static const int myLL = 1;
	static const int myLR = 0;
};

#endif
