/* Copyright (c) 2007 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/** 
 * FVbatterModel.cpp - 2 Dimensional Electrochemical Battery Model main program
 * @author James Geraci
 * @second author Sudarshan Raghunathan
 */

// FVbatterModel.cpp : Defines the entry point for the console application.

#include "StdAfx.h"
#include <iostream>
#include "FV.h"
#include "PV.h"
#include <time.h>

extern spe_program_handle_t spu_prog;

FILE* Vfile;

void writeJtoDisk(double* vectorName, int vectorLength){
  static int count;
  static char file_name[1024];
  snprintf(file_name, 1024, "myJacob_%d.dat", count);
  if((Vfile = fopen(file_name,"w")) == NULL)
    printf("File could not be opened.\n");
  else{
    fwrite((void*)vectorName, sizeof(double), vectorLength,Vfile);
  }
  fflush(Vfile);
  fclose(Vfile);
  count++;
    return;
}


void writeVtoDisk(double* vectorName, int vectorLength){
  static int count;
  static char file_name[1024];
  snprintf(file_name, 1024, "mySV_%d.dat", count);
  if((Vfile = fopen(file_name,"w")) == NULL)
    printf("File could not be opened.\n");
  else{
    fwrite((void*)vectorName, sizeof(double), vectorLength,Vfile);
  }
  fflush(Vfile);
  fclose(Vfile);
  count++;
    return;
}

static double  vector_inf_norm(double* x, int length)
{
  double max, max_position;
  max = fabs(x[0]);
  max_position = 0;
  for(int k = 1; k < length; k++){
    if(fabs(x[k]) > max)
      {
	max = fabs(x[k]);
	max_position = k;
      }
  }
  return max;
}

/**
 *  Replaces x with x - y
 */
static inline void inplace_subtract(double* x, double* y, int length)
{
  for(int i = 0;i < length; x[i] -= y[i], i ++);
}


int main(int argc, char* argv[])
{
  mode = 1; // Want a skyline Jacobian

  double tester = 0;
  int pass = 0;
  int tStep = 0;
  double myLimit = 7.5e-7;// for gmres 1e-8 for \\ 1e-15
  double totalTime = 0;
  
  double* tStepSize = (double*) malloc((sizeof(double)));
  double* tTime = (double*) malloc((sizeof(double)));
  memset(tTime,0,(sizeof(double)));
  memset(tStepSize,0,(sizeof(double)));

  cellSizeData* csData;
  csData = new cellSizeData();
  csData->setnumofElectrodes(3);

  csData->setldxFVColumns(11);
  csData->setElectrolyteFVColumns(11);
  csData->setldFVColumns(11);
  csData->setnumFVRows(31);
  csData->setnumFVZrows(1);

  csData->setldxPVColumns(csData->getldxFVColumns()+1);
  csData->setElectrolytePVColumns(csData->getElectrolyteFVColumns()-1);
  csData->setldPVColumns(csData->getldFVColumns()+1);
	
  double ldxHeight = 10.0;
  double ElectrolyteHeight = 10.0;
  double ldHeight = 10.0;

	
  double ldxDepth = 7.0;
  double ElectrolyteDepth = 7.0;
  double ldDepth = 7.0;

	
  double ldxWidth = 0.03;
  double ElectrolyteWidth = 0.03;
  double ldWidth = 0.03;

  csData->computeTotalFVColumns();
  csData->computeTotalPVColumns();
	
  csData->setnumPVRows(csData->getnumFVRows()+1);
  csData->setnumPVZrows(1);

  csData->computeTotalFVVolumes();
  csData->computeTotalPVVolumes();


  matrixData* myMatrix;
  try{
    myMatrix = new matrixData(csData);
  }
  catch(std::bad_alloc){
    std::cout << "Couldn't Allocate myMatrix!" << std::endl;
    exit(-1);
  }


  FV** myFv;
	
  try{
    myFv = (FV**) new char [(csData->gettotalFVVolumes())*sizeof(FV*)];
  }
  catch(std::bad_alloc){
    std::cout << "Couldn't Allocate myFv array!" << std::endl;
    exit(-1);
  }
	
  for(int counter = 0; counter < csData->gettotalFVVolumes(); counter++)	{
    myFv[counter] = NULL;
  };
	
	
  PV** myPv;
	
  try{
    myPv = (PV**) new char [(csData->gettotalPVVolumes())*sizeof(PV*)];
  }
  catch(std::bad_alloc){
    std::cout << "Couldn't Allocate myFv array!" << std::endl;
    exit(-1);
  }
	
  for(int counter = 0; counter < csData->gettotalPVVolumes(); counter++)
    {
      myPv[counter] = NULL;
    };
	
  electrodeSizeData** electSizeVector;
	
  try{
    electSizeVector = (electrodeSizeData**) new char [csData->getnumofElectrodes()*sizeof(electrodeSizeData*)];
  }
  catch(std::bad_alloc){
    std::cout << "Couldn't Allocate electSizeVector!" << std::endl;
    exit(-1);
  }
	
  try{
    electSizeVector[0] = new electrodeSizeData(ldxWidth,ldxHeight,ldxDepth,
					       csData->getldxFVColumns(),csData->getnumFVRows(),csData->getnumFVZrows(),
					       csData->getldxPVColumns(),csData->getnumPVRows(),csData->getnumPVZrows(),
					       0,0,0,0,csData);
  }
  catch(std::bad_alloc){
    std::cout << "Couldn't Allocate electSizeVector[0]!" << std::endl;
    exit(-1);
  }
  try{
    electSizeVector[1] = new electrodeSizeData(ElectrolyteWidth,ElectrolyteHeight,ElectrolyteDepth,
					       csData->getElectrolyteFVColumns(),csData->getnumFVRows(),csData->getnumFVZrows(),
					       csData->getElectrolytePVColumns(),csData->getnumPVRows(),csData->getnumPVZrows(),
					       0,csData->getldxFVColumns(),0,csData->getldxPVColumns(),csData);
  }
  catch(std::bad_alloc){
    std::cout << "Couldn't Allocate electSizeVector[1]!" << std::endl;
    exit(-1);
  }
  try{
    electSizeVector[2] = new electrodeSizeData(ldWidth,ldHeight,ldDepth,
					       csData->getldFVColumns(),csData->getnumFVRows(),csData->getnumFVZrows(),
					       csData->getldPVColumns(),csData->getnumPVRows(),csData->getnumPVZrows(),
					       0,csData->getldxFVColumns()+csData->getElectrolyteFVColumns(),0,
					       csData->getldxPVColumns()+csData->getElectrolytePVColumns(),csData);
  }
  catch(std::bad_alloc){
    std::cout << "Couldn't Allocate electSizeVector[2]!" << std::endl;
    exit(-1);
  }
	
  Electrode** myElectrodes;
  try{
    myElectrodes = (Electrode**) new char [csData->getnumofElectrodes()*sizeof(Electrode*)];
  }
  catch(std::bad_alloc){
    std::cout << "Couldn't Allocate myElectrodes!" << std::endl;
    exit(-1);
  }

	
  try{
    myElectrodes[0] = new ldxElectrode(myFv, myPv, myMatrix,electSizeVector[0]);
  }
  catch(std::bad_alloc){
    std::cout << "Couldn't Allocate myElectrodes[0] ldxElectrode!" << std::endl;
    exit(-1);
  }
	
  for(int FvCounter = 0; FvCounter < csData->gettotalFVVolumes(); FvCounter++)
    {
      myFv[FvCounter] = myFv[FvCounter];
    }
	
  try{
    myElectrodes[1] = new Electrolyte(myFv, myPv, myMatrix,electSizeVector[1]);
  }
  catch(std::bad_alloc){
    std::cout << "Couldn't Allocate myElectrodes[1] Electrolyte!" << std::endl;
    exit(-1);
  }
	
  for(int FvCounter = 0; FvCounter < csData->gettotalFVVolumes(); FvCounter++)
    {
      myFv[FvCounter] = myFv[FvCounter];
    }
	
  try{
    myElectrodes[2] = new ldElectrode(myFv, myPv, myMatrix,electSizeVector[2]);
  }
  catch(std::bad_alloc){
    std::cout << "Couldn't Allocate myElectrodes[2] ldElectrode!" << std::endl;
    exit(-1);
  }
	
  for(int FvCounter = 0; FvCounter < csData->gettotalFVVolumes(); FvCounter++)
    {
      myFv[FvCounter] = myFv[FvCounter];
    }
	
  for(int FvCounter = 0; FvCounter < csData->gettotalFVVolumes(); FvCounter++)
    {
      myFv[FvCounter]->setupPvPointers();
    }
	
  for(int PvCounter = 0; PvCounter < csData->gettotalPVVolumes(); PvCounter++)
    {
      myPv[PvCounter]->setupFvPointers(myFv);
      myPv[PvCounter]->initializeSize();
      myPv[PvCounter]->initializembScale();
    }

  myMatrix->storesV();

  pass++;

  speid_t ids[NUM_SPU];
  CONTROL_BLOCK_T cb[NUM_SPU] __attribute__((aligned(16)));
  
  //  Get the SPU threads running	
  static bool spu_init;
  
  if (!spu_init) {    
    spu_init = true;
    for (int spu = 0; spu < NUM_SPU; spu ++) {
      // RS: First launch all the threads without the control block,
      // RS: but pass in the logical rank of the thread as the argp
      ids[spu] = spe_create_thread(0, &spu_prog, &(cb[0]), NULL, -1, SPE_MAP_PS);
      // RS: Then fill-in the control-block parameters
      cb[spu].rank             = spu;
      cb[spu].local_store_addr = reinterpret_cast<uint32_t>(spe_get_ls(ids[spu]));
      printf("Obtained a new SPU id: 0x%x, with local_store_addr = 0x%x\n", ids[spu], cb[spu].local_store_addr);
      cb[spu].svLength_size    = myMatrix->svLength; 
      cb[spu].matrixData_addr  = reinterpret_cast<uint32_t>(myMatrix->JacobianStorage);
      cb[spu].rhs_addr         = reinterpret_cast<uint32_t>(myMatrix->rhs);
      cb[spu].hbWidth          = myMatrix->hbWidth;
    }    

    // Wait for all the SPUs to initialize their data
    int sum = 0;
    while(sum != NUM_SPU){
      for(int i = 0; i < NUM_SPU; i++){
        if(spe_stat_out_mbox(ids[i])){
	  sum += spe_read_out_mbox(ids[i]);
        }
        //   printf("PPU sum = %d\n",sum);
      }
    }    
  }

  myMatrix->ids = ids;

  std::cout << "Matrix Data svLength " << myMatrix->svLength << std::endl;  
  while(tStep < tSteps)
    {

      for(int FvolumeCounter = 0; FvolumeCounter < csData->gettotalFVVolumes(); FvolumeCounter++)
	{
	  myFv[FvolumeCounter]->computeDiv();
	  //	myMatrix->cppSVtoMatlab();
	  //	myMatrix->cpfluxVtoMatlab();
	}
      for(int FvolumeCounter = 0; FvolumeCounter < csData->gettotalFVVolumes(); FvolumeCounter++)
	{
	  if(pass==1){
	    myFv[FvolumeCounter]->computeNextEPS();
	    myFv[FvolumeCounter]->computeNextSOC();
	  }
	}
      // Compute the Material Properties
      for(int FvolumeCounter = 0; FvolumeCounter < csData->gettotalFVVolumes(); FvolumeCounter++)
	{
	  myFv[FvolumeCounter]->computeMaterialProperties();
	}

      // Compute the Fluxes
      for(int FvolumeCounter = 0; FvolumeCounter < csData->gettotalFVVolumes(); FvolumeCounter++)
	{
	  myFv[FvolumeCounter]->computeFluxes();
	  //	myMatrix->cppSVtoMatlab();
	  //	myMatrix->cpfluxVtoMatlab();
	}
	    
      myMatrix->resetdivV();
      for(int PvCounter = 0; PvCounter < csData->gettotalPVVolumes(); PvCounter++)
	{ 
	  myPv[PvCounter]->computeDiv();
	}
	    
      // Compute Intermediate value from fluxes
      for(int PvCounter = 0; PvCounter < csData->gettotalPVVolumes(); PvCounter++)
	{ 
	  myPv[PvCounter]->computeDerivatives();
	}

      for(int PvCounter = 0; PvCounter < csData->gettotalPVVolumes(); PvCounter++)
	{
	  //myPv[PvCounter]->computeDerivatives();
	  myPv[PvCounter]->computeIValues();
	  myPv[PvCounter]->setupJacobian();
	}
     
      std::cout << "iV norm = " << vector_inf_norm(myMatrix->myiV(),myMatrix->mysvLength()) << std::endl;
     
      

      //      myMatrix->applyPreConditioner();
      //	 if(pass == 2){
      //cout << "The Jacobian Byte Size is " <<  myMatrix->myJacobSize() << endl;
      // writeJtoDisk(myMatrix->myJacobianStorage(),myMatrix->myJacobSize());	    	    
      // cout << "Waiting after write, press enter to continue" << endl;
      //cin >> myBaby;
      std::cout << "The tStep = " << tStep << " Pass= " << pass << " tester=" << tester << std::endl;
	       //writeVtoDisk(myMatrix->myiV(),myMatrix->mySvByteSize());	    	    
	       //      }    
      myMatrix->applyBoundaryConditions();      


        printf("computing iv\n"); fflush(stdout);
	myMatrix->computeIV();

	inplace_subtract(myMatrix->mysV(), myMatrix->myiV(), myMatrix->mysvLength());
	

	myMatrix->setpiV();	
	
	tester = vector_inf_norm(myMatrix->mypiV(),myMatrix->mysvLength());
	std::cout << tester << std::endl;
	std::cout << vector_inf_norm(myMatrix->mysV(),myMatrix->mysvLength()) << std::endl;

      if(tester <= myLimit)
	{	
	  
	  myMatrix->storesV();
	  writeVtoDisk(myMatrix->mysV(),myMatrix->mysvLength());
	  totalTime = totalTime+
	  tStep++;
	  pass = 0;
	}
	    
      myMatrix->resetiV();
      myMatrix->resetRowBuffer();
      myMatrix->resetJacobian();
      myMatrix->resetFluxes();
      pass++;
    }
	
  free((void*) tStepSize);
  free((void*) tTime);
	
  return 0;
}

