/* Copyright (c) 2007 Massachusetts Institute of Technology
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of
 * this software and associated documentation files (the "Software"), to deal in
 * the Software without restriction, including without limitation the rights to
 * use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of
 * the Software, and to permit persons to whom the Software is furnished to do so,
 * subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS
 * FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR
 * COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
 * IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
/* 
 * @author James Geraci
 * @second author Sudarshan Raghunathan
 * contains all data vectors and jacobian matrix
 *
 */
#include "matrixData.h"
#include "StdAfx.h"
#include <iostream>
#include<sys/time.h>
using namespace std;

static inline double gettime()
{
 timeval t;
 if (gettimeofday(&t, 0) != 0)
   perror("gettimeofday");

 return (t.tv_usec / 1000000.0) + t.tv_sec;
}

matrixData::matrixData(void)
{
}

matrixData::~matrixData(void)
{
	free((void*)fluxV);
	free((void*)divV);
	free((void*)epsV);
	free((void*)socV);
	free((void*)psV);
	free((void*)sV);
	free((void*)savesV);
	free((void*)iV);
	free((void*)piV);
	free((void*)rowBuffer);
	free((void*)extraBuffer1);
	free((void*)intBuffer);
	free((void*)Jacobian);
	free((void*)JacobianStorage);
}


matrixData::matrixData(cellSizeData* csData)
{
  first_call = true;

  ldxPVColumns = csData->getldxPVColumns();
  ElectrolytePVColumns = csData->getElectrolytePVColumns();
  ldPVColumns = csData->getldPVColumns();
  numPVColumns = csData->getnumPVColumns();
  numPVRows = csData->getnumPVRows();
  numPVZrows = csData->getnumPVZrows();
  totalPVvolumes = numPVColumns*numPVRows*numPVZrows;

  numFVColumns = csData->getnumFVColumns();
  numFVRows = csData->getnumFVRows();
  numFVZrows = csData->getnumFVZrows();
  totalFVvolumes = numFVColumns*numFVRows*numFVZrows;
  
  svLength = totalPVvolumes*myVars;
  svByteSize = svLength*sizeof(double);
  
  cout << "Original svByteSize " << svByteSize << endl;
  
  JacobSize = svLength*svLength;
  JacobByteSize = JacobSize*sizeof(double);
  FVsvLength = totalFVvolumes*(myVars+2);
  FVsvByteSize = FVsvLength*sizeof(double);
  // Compute Number of Variables in Sparse Matrix
  int ADEGH = 5*totalPVvolumes - 2*(totalPVvolumes/numPVRows) - 2*(numPVRows - 1) - 2;
  int R = totalPVvolumes - numPVRows*(ElectrolytePVColumns);
  int F = ADEGH - 2*numPVRows - 5*ElectrolytePVColumns*numPVRows + 2*(ElectrolytePVColumns+1);
  int B = ADEGH - 5*(ElectrolytePVColumns + 1)*numPVRows + 2*(ElectrolytePVColumns + 1);

  hbWidth = (numPVColumns+1)*myVars;
  hbWbytes = hbWidth*sizeof(double);
  ActualsparseMatrixSize = 5*ADEGH + R + F + B;
  int maxnzPerRow = 16;
  sparseMatrixSize = myVars*maxnzPerRow*totalPVvolumes; 
  f = fluxesPerVar*FVsvByteSize;
  try{
    val_CRS = (double*) malloc_align(sparseMatrixSize*sizeof(double),7);
    row_ptr = (double*) malloc_align(sparseMatrixSize*sizeof(double),7);
    col_ind = (double*) malloc_align(sparseMatrixSize*sizeof(double),7);
    nzInRow = (unsigned int*) malloc_align(svLength*sizeof(unsigned int),7);
    
    rhs = (double*) malloc_align (svByteSize * 2, 7);
    socV = (double*) malloc_align(totalFVvolumes*sizeof(double),7);
    fluxV = (double*) malloc_align(f,7);
    divV = (double*) malloc_align(totalPVvolumes*sizeof(double),7);
    epsV = (double*) malloc_align(totalFVvolumes*sizeof(double),7);
    // can put divV here too.
    psV = (double*) malloc_align(svByteSize,7);
    sV = (double*) malloc_align(svByteSize,7);
    savesV = (double*) malloc_align(svByteSize,7);
    iV = (double*) malloc_align(svByteSize,7);
    piV = (double*) malloc_align(svByteSize,7);
    rowBuffer = (double*) malloc_align(svByteSize,7);
    extraBuffer1 = (double*) malloc_align(svByteSize,7);
    intBuffer = (int*) malloc_align(svLength*sizeof(int),7);
    //spu_rhs = reinterpret_cast<double*>(malloc_align(svByteSize * 2, 7));
    Jacobian = (double**) malloc_align(svLength*sizeof(double*),7);
    JacobianStorage = (double*) malloc_align(JacobByteSize,7);
 resetJacobian(); 
   for(int i =0; i < svLength; i++){
      Jacobian[i] = &JacobianStorage[i*svLength];
    }
  }
  catch(std::bad_alloc)
    {
      cout << "Aligned Malloc_Align has Failed!" << endl;
      exit(-1);
	}
  for(int d = 0; d < sparseMatrixSize; d++)
    {
      val_CRS[d] = 0;
      row_ptr[d] = -1;
      col_ind[d] = -1;
    }
  memset(fluxV,0,f);
  memset(socV,0,totalFVvolumes*sizeof(double));
  memset(divV,0,totalPVvolumes*sizeof(double));
  memset(epsV,0,totalFVvolumes*sizeof(double));
  memset(piV,0,svByteSize);
  memset(iV,0,svByteSize);
  memset(sV,0,svByteSize);
  memset(psV,0,svByteSize);
  memset(extraBuffer1,0,svByteSize);
  memset(savesV,0,svByteSize);
  memset(rowBuffer,0,svByteSize);
  memset(rhs,0,2*svByteSize);

  printf("Matrix construction completed.\n");
}

void matrixData::computeIV(){
  double et_start,et_end,et_total;
  double et_start_rowx,et_end_rowx,et_total_rowx;
  double et_start_bs,et_end_bs,et_total_bs;
  et_total = 0;
  et_total_rowx = 0;
  et_total_bs = 0;
  tic = gettime();
// wait for the spus to sync
  first_call = false;
  sum = 0;
  present_base_row = 0;
  for(int j = 0; j < svLength; j++){
    rhs[2*j] = iV[j];
    rhs[2*j+1] = iV[j];
    
  }
  
  printf("Going to unstall the SPUs\n");
   
  for(int spu = 0; spu < NUM_SPU; spu ++)
    spe_write_in_mbox(ids[spu], spu);
  
    // Where are doing the solve for the first base row here

  while(present_base_row < svLength-1) {
    //  printf("present_base_row = %d\n", present_base_row); fflush(stdout);
     et_start = gettime(); 
    while(sum != NUM_SPU){
      for(int i = 0; i < NUM_SPU; i++){
	if(spe_stat_out_mbox(ids[i])){
	  sum += spe_read_out_mbox(ids[i]);
	}
      }
    }
    et_end = gettime();
    et_total += et_end-et_start;

    // Want to see time  spent finding row exchanges
    et_start_rowx = gettime();
    sum = 0;
    present_base_row ++;
    // Here is where we do the partial pivoting
    max_pivot_value = fabs(Jacobian[present_base_row][present_base_row]);
    max_pivot = present_base_row;
    // Start Partial Pivoting here
    /*    for(int i = present_base_row + 1;  i < min(svLength,present_base_row+hbWidth); i++){
      fmax = fabs(Jacobian[i][present_base_row]); 
      if(fmax > max_pivot_value){
	max_pivot_value = fmax;
	max_pivot = i;
      }
    }
    */
    // max_pivot = min(present_base_row + 5, svLength-1);
    // max_pivot = present_base_row + 5;
    // Exchange the rows in the Jacobian 
    if(max_pivot != present_base_row && max_pivot < min(svLength,present_base_row+hbWidth)){
      bufaddrf =  Jacobian[max_pivot];
      Jacobian[max_pivot] =  Jacobian[present_base_row];
      Jacobian[present_base_row] =  bufaddrf;
      rhsBuffer = rhs[2*max_pivot];
      rhs[2*max_pivot] = rhs[2*present_base_row];
      rhs[2*present_base_row] = rhsBuffer;
    }
    // End Partial Pivot Here
    for(int spu = 0; spu < NUM_SPU; spu ++)
      spe_write_in_mbox(ids[spu], max_pivot);    
    et_end_rowx = gettime();
    et_total_rowx += et_end_rowx-et_start_rowx;
    
  }
  
  // }
  
  //printf("All SPUs are done, back-solving\n");
  et_start_bs = gettime();
   for(int j = 0; j < svLength; j++){
    iV[j] = rhs[2*j];
  }
  
  for (int j = svLength - 1; j >= 0; j --) {
    for (int k = j + 1; k < min(svLength,j+hbWidth); k ++) {
      iV[j] -= Jacobian[j][k] * iV[k];
    }
    iV[j] = iV[j]/Jacobian[j][j];
    }
  et_end_bs = gettime();
  et_total_bs = et_end_bs - et_start_bs;
  printf("Finished Newton iteration.\n"); fflush(stdout);
  sum = 0;
  present_base_row = 0;
  toc = gettime();
  elapsed = toc-tic;
  printf("Elased Time = %f\n",elapsed);
  // printf("Elased Time = %f\n",elapsed);
  printf("Total time Syncing Rows = %f and Compute time = %f total back solve time = %f\n",et_total_rowx, et_total, et_total_bs);
			     
}

int matrixData::myPVColumns()
{
  return numPVColumns;
};

int matrixData::myPVRows()
{
  return numPVRows;
};

int matrixData::myPVZrows()
{
  return numPVZrows;
};

int matrixData::myFVColumns()
{
  return numFVColumns;
};

int matrixData::myFVRows()
{
  return numFVRows;
};

int matrixData::myFVZrows()
{
  return numFVZrows;
};


double** matrixData::myJacobian()
{
  return Jacobian;
};

double* matrixData::myfluxV(){
  return fluxV;
};

double* matrixData::mydivV()
{
  return divV;
};

double* matrixData::myepsV()
{
  return epsV;
};

double* matrixData::mysocV()
{
  return socV;
};

double* matrixData::myiV()
{
  return iV;
};

double* matrixData::mypiV()
{
  return piV;
};

double* matrixData::mysV()
{
  return sV;
};

double* matrixData::mypsV()
{
  return psV;
};

int matrixData::myJacobByteSize()
{
  return JacobByteSize;
};

int matrixData::myJacobSize()
{
  return JacobSize;
};

int matrixData::mysvLength()
{
  return svLength;
};

void matrixData::setpiV()
{
  for(int counter = 0; counter < svLength; counter++)
    {
	piV[counter] = fabs(iV[counter]);
    }
  return;
};

void matrixData::resetiV()
{
  memset(iV,0,svByteSize);
  return;
}

void matrixData::resetdivV()
{
  memset(divV,0,totalPVvolumes*sizeof(double));
  return;
}

void matrixData::resetFluxes()
{
  memset(fluxV,0,f);
  return;
}

double* matrixData::myJacobianStorage(){
  return JacobianStorage;
}

void matrixData::resetJacobian()
{
  memset(JacobianStorage,0,JacobByteSize);
  return;
};

void matrixData::writeVtoDisk(double* vectorName, int vectorLength){
  if((Vfile = fopen("myVector.dat","w")) == NULL)
    printf("File could not be opened.\n");
  else{
    fwrite((void*)vectorName, vectorLength*sizeof(double), vectorLength,Vfile);
  }
  fclose(Vfile);
    return;
}


void matrixData::restoresV()
{
  memcpy((void*)sV,(void *) psV, svByteSize); 
  return;
};


void matrixData::storesV()
{
  memcpy((void*)psV,(void *) sV, svByteSize);
  return;
};

void matrixData::resetRowBuffer()
{
  memset(rowBuffer,0,svByteSize);
  return;
};

double* matrixData::myRowBuffer()
{
  return rowBuffer;
};

int matrixData::mySvByteSize()
{
  return svByteSize;
};


double* matrixData::myExtraBuffer1()
{
  return extraBuffer1;
};

int* matrixData::myintBuffer()
{
  return intBuffer;
};

/*******************************************************/
/*********** Boundary Conditions for Row Major *********/
/*******************************************************/
 void matrixData::applyBoundaryConditions(void){
  int clearRow;
  
  // EK Rows
	
  for(int j = 0; j < numPVRows; j++){
  for(int i = 0; i < ElectrolytePVColumns; i++){
    if(mode == 0){
      clearRow = (totalPVvolumes*2+ldxPVColumns + i + j*numPVColumns);
    }
    else{
      clearRow = myVars*ldxPVColumns + myVars*i + 2 + j*myVars*numPVColumns;
    }
//  clearRow = (totalPVvolumes*2+ldxPVColumns + i + j*numPVColumns);
  iV[clearRow] = 0;
  //	Jacobian[clearRow] = 1e5;
  memset((void*) &Jacobian[clearRow][0],0,svByteSize);
  Jacobian[clearRow][clearRow] = 1;
  }
  }
  
  // DOC Row
  if(mode == 0){
    clearRow = (totalPVvolumes*2- 1);
  }else{
    clearRow = svLength-2;
  }
  iV[clearRow] = 0;
  memset((void*) &Jacobian[clearRow][0],0,svByteSize);
  for(int j = 0; j < svLength; j++)
  {
  Jacobian[clearRow][j] = 0;
  }
  Jacobian[clearRow][clearRow] = 1;
  return;
  };



static double  vector_inf_norm(double* x, int length)
{
  double max, max_position;
  max = fabs(x[0]);
  max_position = 0;
  for(int k = 1; k < length; k++){
    if(fabs(x[k]) > max)
      {
	max = fabs(x[k]);
	max_position = k;
      }
  }
  return max;
}

/*******************************************************/
/*********** PreConditioner for Row Major **************/
/*******************************************************/
void matrixData::applyPreConditioner(void){
  double myMax;
  
  int end = totalPVvolumes*2;
  for(int index = totalPVvolumes; index < end; index++){
    for(int r = 0; r < svLength; r++){
      extraBuffer1[r] = Jacobian[index][r];
      //	cout << extraBuffer1[r] << endl;
    }
    
    myMax = vector_inf_norm(extraBuffer1, svLength);
    //	cout << myMax;
    myMax = sqrt(fabs(myMax));
    for(int p = 0; p < svLength; p++){
      Jacobian[index][p] = Jacobian[index][p]/myMax;
    }
    iV[index] = iV[index]/myMax;
  }
  return;
};
